@@ -118,7 +118,7 @@ static void deserialize_stable_size(MVMThreadContext *tc, MVMSTable *st, MVMSeri
 /* Serializes the REPR data. */
 static void serialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerializationWriter *writer) {
     MVMP6numREPRData *repr_data = (MVMP6numREPRData *)st->REPR_data;
-    MVM_serialization_write_varint(tc, writer, repr_data->bits);
+    MVM_serialization_write_int(tc, writer, repr_data->bits);
 }
 
 /* Deserializes representation data. */
@@ -126,7 +126,7 @@ static void deserialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerial
     MVMP6numREPRData *repr_data = (MVMP6numREPRData *)MVM_malloc(sizeof(MVMP6numREPRData));
 
 
-    repr_data->bits        = MVM_serialization_read_varint(tc, reader);
+    repr_data->bits        = MVM_serialization_read_int(tc, reader);
 
     if (repr_data->bits !=  1 && repr_data->bits !=  2 && repr_data->bits !=  4 && repr_data->bits != 8
      && repr_data->bits != 16 && repr_data->bits != 32 && repr_data->bits != 64)
@@ -187,7 +187,6 @@ static const MVMREPROps this_repr = {
     NULL, /* spesh */
     "P6num", /* name */
     MVM_REPR_ID_P6num,
-    0, /* refs_frames */
     NULL, /* unmanaged_size */
     NULL, /* describe_refs */
 };
