@@ -1,5 +1,8 @@
 #include "moar.h"
 
+/* This allows the dynlex cache to be disabled when bug hunting, if needed. */
+#define MVM_DYNLEX_CACHE_ENABLED 1
+
 /* Computes the initial work area for a frame or a specialization of a frame. */
 MVMRegister * MVM_frame_initial_work(MVMThreadContext *tc, MVMuint16 *local_types,
                                      MVMuint16 num_locals) {
@@ -132,22 +135,31 @@ static MVMFrame * create_context_only(MVMThreadContext *tc, MVMStaticFrame *stat
         frame->env = MVM_fixed_size_alloc(tc, tc->instance->fsa, static_frame->body.env_size);
         frame->allocd_env = static_frame->body.env_size;
         if (autoclose) {
-            MVMuint16 i;
-            for (i = 0; i < static_frame->body.num_lexicals; i++) {
+            MVMuint16 i, num_lexicals = static_frame->body.num_lexicals;
+
+            MVM_gc_root_temp_push(tc, (MVMCollectable **)&static_frame);
+
+            for (i = 0; i < num_lexicals; i++) {
                 if (!static_frame->body.static_env[i].o && static_frame->body.static_env_flags[i] == 1) {
                     MVMint32 scid, objid;
                     if (MVM_bytecode_find_static_lexical_scref(tc, static_frame->body.cu,
                             static_frame, i, &scid, &objid)) {
+                        MVMObject *resolved;
                         MVMSerializationContext *sc = MVM_sc_get_sc(tc, static_frame->body.cu, scid);
+
                         if (sc == NULL)
                             MVM_exception_throw_adhoc(tc,
                                 "SC not yet resolved; lookup failed");
+
+                        resolved = MVM_sc_get_object(tc, sc, objid);
+
                         MVM_ASSIGN_REF(tc, &(static_frame->common.header),
                             static_frame->body.static_env[i].o,
-                            MVM_sc_get_object(tc, sc, objid));
+                            resolved);
                     }
                 }
             }
+            MVM_gc_root_temp_pop(tc);
         }
         memcpy(frame->env, static_frame->body.static_env, static_frame->body.env_size);
     }
@@ -1002,16 +1014,23 @@ void MVM_frame_unwind_to(MVMThreadContext *tc, MVMFrame *frame, MVMuint8 *abs_ad
 MVMObject * MVM_frame_get_code_object(MVMThreadContext *tc, MVMCode *code) {
     if (REPR(code)->ID != MVM_REPR_ID_MVMCode)
         MVM_exception_throw_adhoc(tc, "getcodeobj needs a code ref");
+
     if (!code->body.code_object) {
         MVMStaticFrame *sf = code->body.sf;
         if (sf->body.code_obj_sc_dep_idx > 0) {
+            MVMObject *resolved;
             MVMSerializationContext *sc = MVM_sc_get_sc(tc, sf->body.cu,
                 sf->body.code_obj_sc_dep_idx - 1);
             if (sc == NULL)
                 MVM_exception_throw_adhoc(tc,
                     "SC not yet resolved; lookup failed");
+
+            MVMROOT(tc, code, {
+                resolved = MVM_sc_get_object(tc, sc, sf->body.code_obj_sc_idx);
+            });
+
             MVM_ASSIGN_REF(tc, &(code->common.header), code->body.code_object,
-                MVM_sc_get_object(tc, sc, sf->body.code_obj_sc_idx));
+                resolved);
         }
     }
     return code->body.code_object;
@@ -1030,6 +1049,34 @@ void MVM_frame_capturelex(MVMThreadContext *tc, MVMObject *code) {
     MVM_ASSIGN_REF(tc, &(code->header), code_obj->body.outer, captured);
 }
 
+/* This is used for situations in Perl 6 like:
+ * supply {
+ *     my $x = something();
+ *     whenever $supply {
+ *         QUIT { $x.foo() }
+ *     }
+ * }
+ * Here, the QUIT may be called without an invocation of the whenever ever
+ * having taken place. At the point we closure-clone the whenever block, we
+ * will capture_inner the QUIT phaser. This creates a fake outer for the
+ * QUIT, but makes *its* outer point to the nearest instance of the relevant
+ * static frame on the call stack, so that the QUIT will disocver the correct
+ * $x.
+ */
+void MVM_frame_capture_inner(MVMThreadContext *tc, MVMObject *code) {
+    MVMCode *code_obj = (MVMCode *)code;
+    MVMFrame *outer;
+    MVMROOT(tc, code, {
+        MVMStaticFrame *sf_outer = code_obj->body.sf->body.outer;
+        outer = create_context_only(tc, sf_outer, (MVMObject *)sf_outer->body.static_code, 1);
+        MVMROOT(tc, outer, {
+            MVMFrame *outer_outer = autoclose(tc, sf_outer->body.outer);
+            MVM_ASSIGN_REF(tc, &(outer->header), outer->outer, outer_outer);
+        });
+    });
+    MVM_ASSIGN_REF(tc, &(code->header), code_obj->body.outer, outer);
+}
+
 /* Given the specified code object, copies it and returns a copy which
  * captures a closure over the current scope. */
 MVMObject * MVM_frame_takeclosure(MVMThreadContext *tc, MVMObject *code) {
@@ -1089,17 +1136,20 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
     }
     flag  = flags ? flags[effective_idx] : -1;
     if (flag != -1 && static_env[effective_idx].o == NULL) {
-        MVMStaticFrameBody *static_frame_body = &(f->static_info->body);
         MVMint32 scid, objid;
         if (MVM_bytecode_find_static_lexical_scref(tc, effective_sf->body.cu,
                 effective_sf, effective_idx, &scid, &objid)) {
             MVMSerializationContext *sc = MVM_sc_get_sc(tc, effective_sf->body.cu, scid);
+            MVMObject *resolved;
             if (sc == NULL)
                 MVM_exception_throw_adhoc(tc,
                     "SC not yet resolved; lookup failed");
+            MVMROOT(tc, f, {
+                resolved = MVM_sc_get_object(tc, sc, objid);
+            });
             MVM_ASSIGN_REF(tc, &(f->static_info->common.header),
-                static_frame_body->static_env[effective_idx].o,
-                MVM_sc_get_object(tc, sc, objid));
+                f->static_info->body.static_env[effective_idx].o,
+                resolved);
         }
     }
     if (flag == 0) {
@@ -1128,15 +1178,12 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
  * if it does not exist. Incorrect type always throws. */
 MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 type) {
     MVMFrame *cur_frame = tc->cur_frame;
-    MVM_string_flatten(tc, name);
     while (cur_frame != NULL) {
         MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
         if (lexical_names) {
             /* Indexes were formerly stored off-by-one to avoid semi-predicate issue. */
             MVMLexicalRegistry *entry;
-
             MVM_HASH_GET(tc, lexical_names, name, entry)
-
             if (entry) {
                 if (cur_frame->static_info->body.lexical_types[entry->value] == type) {
                     MVMRegister *result = &cur_frame->env[entry->value];
@@ -1168,7 +1215,6 @@ MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *na
  * chain. */
 MVM_PUBLIC void MVM_frame_bind_lexical_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 type, MVMRegister *value) {
     MVMFrame *cur_frame = tc->cur_frame;
-    MVM_string_flatten(tc, name);
     while (cur_frame != NULL) {
         MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
         if (lexical_names) {
@@ -1220,15 +1266,12 @@ MVMObject * MVM_frame_find_lexical_by_name_outer(MVMThreadContext *tc, MVMString
 /* Looks up the address of the lexical with the specified name, starting with
  * the specified frame. Only works if it's an object lexical.  */
 MVMRegister * MVM_frame_find_lexical_by_name_rel(MVMThreadContext *tc, MVMString *name, MVMFrame *cur_frame) {
-    MVM_string_flatten(tc, name);
     while (cur_frame != NULL) {
         MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
         if (lexical_names) {
             /* Indexes were formerly stored off-by-one to avoid semi-predicate issue. */
             MVMLexicalRegistry *entry;
-
             MVM_HASH_GET(tc, lexical_names, name, entry)
-
             if (entry) {
                 if (cur_frame->static_info->body.lexical_types[entry->value] == MVM_reg_obj) {
                     MVMRegister *result = &cur_frame->env[entry->value];
@@ -1253,7 +1296,6 @@ MVMRegister * MVM_frame_find_lexical_by_name_rel(MVMThreadContext *tc, MVMString
 /* Looks up the address of the lexical with the specified name, starting with
  * the specified frame. It checks all outer frames of the caller frame chain.  */
 MVMRegister * MVM_frame_find_lexical_by_name_rel_caller(MVMThreadContext *tc, MVMString *name, MVMFrame *cur_caller_frame) {
-    MVM_string_flatten(tc, name);
     while (cur_caller_frame != NULL) {
         MVMFrame *cur_frame = cur_caller_frame;
         while (cur_frame != NULL) {
@@ -1261,9 +1303,7 @@ MVMRegister * MVM_frame_find_lexical_by_name_rel_caller(MVMThreadContext *tc, MV
             if (lexical_names) {
                 /* Indexes were formerly stored off-by-one to avoid semi-predicate issue. */
                 MVMLexicalRegistry *entry;
-
                 MVM_HASH_GET(tc, lexical_names, name, entry)
-
                 if (entry) {
                     if (cur_frame->static_info->body.lexical_types[entry->value] == MVM_reg_obj) {
                         MVMRegister *result = &cur_frame->env[entry->value];
@@ -1290,6 +1330,7 @@ MVMRegister * MVM_frame_find_lexical_by_name_rel_caller(MVMThreadContext *tc, MV
 /* Looks up the address of the lexical with the specified name and the
  * specified type. Returns null if it does not exist. */
 static void try_cache_dynlex(MVMThreadContext *tc, MVMFrame *from, MVMFrame *to, MVMString *name, MVMRegister *reg, MVMuint16 type, MVMuint32 fcost, MVMuint32 icost) {
+#if MVM_DYNLEX_CACHE_ENABLED
     MVMint32 next = 0;
     MVMint32 frames = 0;
     MVMuint32 desperation = 0;
@@ -1316,6 +1357,7 @@ static void try_cache_dynlex(MVMThreadContext *tc, MVMFrame *from, MVMFrame *to,
         }
         from = from->caller;
     }
+#endif
 }
 MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 *type, MVMFrame *cur_frame, MVMint32 vivify, MVMFrame **found_frame) {
     FILE *dlog = tc->instance->dynvar_log_fh;
@@ -1336,7 +1378,6 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
         last_time = tc->instance->dynvar_log_lasttime;
     }
 
-    MVM_string_flatten(tc, name);
     while (cur_frame != NULL) {
         MVMLexicalRegistry *lexical_names;
         MVMSpeshCandidate  *cand     = cur_frame->spesh_cand;
@@ -1579,7 +1620,6 @@ MVMRegister * MVM_frame_lexical(MVMThreadContext *tc, MVMFrame *f, MVMString *na
     MVMLexicalRegistry *lexical_names = f->static_info->body.lexical_names;
     if (lexical_names) {
         MVMLexicalRegistry *entry;
-        MVM_string_flatten(tc, name);
         MVM_HASH_GET(tc, lexical_names, name, entry)
         if (entry)
             return &f->env[entry->value];
@@ -1598,7 +1638,6 @@ MVMRegister * MVM_frame_try_get_lexical(MVMThreadContext *tc, MVMFrame *f, MVMSt
     MVMLexicalRegistry *lexical_names = f->static_info->body.lexical_names;
     if (lexical_names) {
         MVMLexicalRegistry *entry;
-        MVM_string_flatten(tc, name);
         MVM_HASH_GET(tc, lexical_names, name, entry)
         if (entry && f->static_info->body.lexical_types[entry->value] == type) {
             MVMRegister *result = &f->env[entry->value];
@@ -1615,7 +1654,6 @@ MVMuint16 MVM_frame_lexical_primspec(MVMThreadContext *tc, MVMFrame *f, MVMStrin
     MVMLexicalRegistry *lexical_names = f->static_info->body.lexical_names;
     if (lexical_names) {
         MVMLexicalRegistry *entry;
-        MVM_string_flatten(tc, name);
         MVM_HASH_GET(tc, lexical_names, name, entry)
         if (entry) {
             switch (f->static_info->body.lexical_types[entry->value]) {
