@@ -12,7 +12,8 @@ void MVM_profile_heap_start(MVMThreadContext *tc, MVMObject *config) {
 
 /* Grows storage if it's full, zeroing the extension. Assumes it's only being
  * grown for one more item. */
-static void grow_storage(void **store, MVMuint64 *num, MVMuint64 *alloc, size_t size) {
+static void grow_storage(void *store_ptr, MVMuint64 *num, MVMuint64 *alloc, size_t size) {
+    void **store = (void **)store_ptr;
     if (*num == *alloc) {
         *alloc = *alloc ? 2 * *alloc : 32;
         *store = MVM_realloc(*store, *alloc * size);
@@ -173,6 +174,8 @@ static MVMuint64 get_collectable_idx(MVMThreadContext *tc,
             idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_STABLE, collectable);
         else if (collectable->flags & MVM_CF_TYPE_OBJECT)
             idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_TYPE_OBJECT, collectable);
+        else if (collectable->flags & MVM_CF_FRAME)
+            idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_FRAME, collectable);
         else
             idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_OBJECT, collectable);
         saw(tc, ss, collectable, idx);
@@ -240,10 +243,13 @@ static void set_static_frame_index(MVMThreadContext *tc, MVMHeapSnapshotState *s
         s = &(ss->col->static_frames[i]);
         if (s->name == name_idx && s->cuid == cuid_idx && s->line == line && s->file == file_idx) {
             col->type_or_frame_index = i;
+            MVM_free(ann);
             return;
         }
     }
 
+    MVM_free(ann);
+
     grow_storage(&(ss->col->static_frames), &(ss->col->num_static_frames),
         &(ss->col->alloc_static_frames), sizeof(MVMHeapSnapshotStaticFrame));
     s = &(ss->col->static_frames[ss->col->num_static_frames]);
@@ -258,7 +264,7 @@ static void set_static_frame_index(MVMThreadContext *tc, MVMHeapSnapshotState *s
 /* Processes the work items, until we've none left. */
 static void process_collectable(MVMThreadContext *tc, MVMHeapSnapshotState *ss,
         MVMHeapSnapshotCollectable *col, MVMCollectable *c) {
-    MVMuint32 sc_idx = MVM_get_idx_of_sc(c);
+    MVMuint32 sc_idx = MVM_sc_get_idx_of_sc(c);
     if (sc_idx > 0)
         add_reference_const_cstr(tc, ss, "<SC>",
             get_collectable_idx(tc, ss,
@@ -267,7 +273,6 @@ static void process_collectable(MVMThreadContext *tc, MVMHeapSnapshotState *ss,
 }
 static void process_gc_worklist(MVMThreadContext *tc, MVMHeapSnapshotState *ss, char *desc) {
     MVMCollectable **c_ptr;
-    MVMFrame *f;
     MVMuint16 ref_kind = desc
         ? MVM_SNAPSHOT_REF_KIND_STRING
         : MVM_SNAPSHOT_REF_KIND_UNKNOWN;
@@ -280,11 +285,6 @@ static void process_gc_worklist(MVMThreadContext *tc, MVMHeapSnapshotState *ss,
             add_reference(tc, ss, ref_kind, ref_index,
                 get_collectable_idx(tc, ss, c));
     }
-    while (f = MVM_gc_worklist_get_frame(tc, ss->gcwl)) {
-        if (f)
-            add_reference(tc, ss, ref_kind, ref_index,
-                get_frame_idx(tc, ss, f));
-    }
 }
 static void process_object(MVMThreadContext *tc, MVMHeapSnapshotState *ss,
         MVMHeapSnapshotCollectable *col, MVMObject *obj) {
@@ -407,7 +407,7 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
                 col.collectable_size = sizeof(MVMFrame);
                 set_static_frame_index(tc, ss, &col, frame->static_info);
 
-                if (frame->caller)
+                if (frame->caller && !MVM_FRAME_IS_ON_CALLSTACK(tc, frame))
                     add_reference_const_cstr(tc, ss, "Caller",
                         get_frame_idx(tc, ss, frame->caller));
                 if (frame->outer)
@@ -449,8 +449,6 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
                     (MVMCollectable *)frame->code_ref, "Code reference");
                 MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
                     (MVMCollectable *)frame->static_info, "Static frame");
-                MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                    (MVMCollectable *)frame->context_object, "Context object");
                 MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
                     (MVMCollectable *)frame->dynlex_cache_name,
                     "Dynamic lexical cache name");
@@ -484,8 +482,9 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
             case MVM_SNAPSHOT_COL_KIND_THREAD_ROOTS: {
                 MVMThreadContext *thread_tc = (MVMThreadContext *)item.target;
                 MVM_gc_root_add_tc_roots_to_worklist(thread_tc, NULL, ss);
-                add_reference_const_cstr(tc, ss, "Current frame",
-                    get_frame_idx(tc, ss, thread_tc->cur_frame));
+                if (!MVM_FRAME_IS_ON_CALLSTACK(tc, tc->cur_frame))
+                    add_reference_const_cstr(tc, ss, "Current frame",
+                        get_frame_idx(tc, ss, thread_tc->cur_frame));
                  break;
             }
             case MVM_SNAPSHOT_COL_KIND_ROOT: {
@@ -511,15 +510,36 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
                             push_workitem(tc, ss,
                                 MVM_SNAPSHOT_COL_KIND_INTERGEN_ROOTS,
                                 cur_thread->body.tc));
+                        add_reference_const_cstr(tc, ss, "Thread Call Stack Roots",
+                            push_workitem(tc, ss,
+                                MVM_SNAPSHOT_COL_KIND_CALLSTACK_ROOTS,
+                                cur_thread->body.tc));
                     }
                     cur_thread = cur_thread->body.next;
                 }
 
                 break;
             }
-            case MVM_SNAPSHOT_COL_KIND_INTERGEN_ROOTS:
-                MVM_gc_root_add_gen2s_to_snapshot(tc, ss);
+            case MVM_SNAPSHOT_COL_KIND_INTERGEN_ROOTS: {
+                MVMThreadContext *thread_tc = (MVMThreadContext *)item.target;
+                MVM_gc_root_add_gen2s_to_snapshot(thread_tc, ss);
                 break;
+            }
+            case MVM_SNAPSHOT_COL_KIND_CALLSTACK_ROOTS: {
+                MVMThreadContext *thread_tc = (MVMThreadContext *)item.target;
+                if (thread_tc->cur_frame && MVM_FRAME_IS_ON_CALLSTACK(tc, thread_tc->cur_frame)) {
+                    MVMFrame *cur_frame = thread_tc->cur_frame;
+                    MVMint32 idx = 0;
+                    while (cur_frame && MVM_FRAME_IS_ON_CALLSTACK(tc, cur_frame)) {
+                        add_reference_idx(tc, ss, idx,
+                            push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_FRAME,
+                                (MVMCollectable *)cur_frame));
+                        idx++;
+                        cur_frame = cur_frame->caller;
+                    }
+                }
+                break;
+            }
             default:
                 MVM_panic(1, "Unknown heap snapshot worklist item kind %d", item.kind);
         }
@@ -561,21 +581,17 @@ void MVM_profile_heap_add_collectable_rel_idx(MVMThreadContext *tc,
 
 /* Drives the overall process of recording a snapshot of the heap. */
 static void record_snapshot(MVMThreadContext *tc, MVMHeapSnapshotCollection *col, MVMHeapSnapshot *hs) {
-    MVMuint64 perm_root_synth;
-
-    /* Iinitialize state for taking a snapshot. */
+    /* Initialize state for taking a snapshot. */
     MVMHeapSnapshotState ss;
     memset(&ss, 0, sizeof(MVMHeapSnapshotState));
     ss.col = col;
     ss.hs = hs;
-    ss.gcwl = MVM_gc_worklist_create(tc, 1, 1);
+    ss.gcwl = MVM_gc_worklist_create(tc, 1);
 
     /* We push the ultimate "root of roots" onto the worklist to get things
      * going, then set off on our merry way. */
-    printf("Recording heap snapshot\n");
     push_workitem(tc, &ss, MVM_SNAPSHOT_COL_KIND_ROOT, NULL);
     process_workitems(tc, &ss);
-    printf("Recording completed\n");
 
     /* Clean up temporary state. */
     MVM_free(ss.workitems);
@@ -643,7 +659,7 @@ MVMObject * types_str(MVMThreadContext *tc, MVMHeapSnapshotCollection *col) {
      MVMuint64 i;
      for (i = 0; i < col->num_types; i++) {
          char tmp[256];
-         size_t item_chars = snprintf(tmp, 256,
+         int item_chars = snprintf(tmp, 256,
             "%"PRIu64",%"PRIu64";",
             col->types[i].repr_name,
             col->types[i].type_name);
@@ -677,7 +693,7 @@ MVMObject * static_frames_str(MVMThreadContext *tc, MVMHeapSnapshotCollection *c
      MVMuint64 i;
      for (i = 0; i < col->num_static_frames; i++) {
          char tmp[256];
-         size_t item_chars = snprintf(tmp, 256,
+         int item_chars = snprintf(tmp, 256,
             "%"PRId64",%"PRId64",%"PRId64",%"PRId64";",
             col->static_frames[i].name,
             col->static_frames[i].cuid,
@@ -713,7 +729,7 @@ MVMObject * collectables_str(MVMThreadContext *tc, MVMHeapSnapshot *s) {
      MVMuint64 i;
      for (i = 0; i < s->num_collectables; i++) {
          char tmp[256];
-         size_t item_chars = snprintf(tmp, 256,
+         int item_chars = snprintf(tmp, 256,
             "%"PRIu16",%"PRId32",%"PRIu16",%"PRIu64",%"PRIu64",%"PRIu32";",
             s->collectables[i].kind,
             s->collectables[i].type_or_frame_index,
@@ -751,7 +767,7 @@ MVMObject * references_str(MVMThreadContext *tc, MVMHeapSnapshot *s) {
     MVMuint64 i;
     for (i = 0; i < s->num_references; i++) {
         char tmp[128];
-        size_t item_chars = snprintf(tmp, 128, "%lu,%lu,%lu;",
+        int item_chars = snprintf(tmp, 128, "%lu,%lu,%lu;",
             s->references[i].description & ((1 << MVM_SNAPSHOT_REF_KIND_BITS) - 1),
             s->references[i].description >> MVM_SNAPSHOT_REF_KIND_BITS,
             s->references[i].collectable_index);
@@ -813,7 +829,13 @@ MVMObject * collection_to_mvm_objects(MVMThreadContext *tc, MVMHeapSnapshotColle
 
 /* Finishes heap profiling, getting the data. */
 MVMObject * MVM_profile_heap_end(MVMThreadContext *tc) {
-    MVMObject *dataset = collection_to_mvm_objects(tc, tc->instance->heap_snapshots);
+    MVMObject *dataset;
+
+    /* Trigger a GC run, to ensure we get at least one heap snapshot. */
+    MVM_gc_enter_from_allocator(tc);
+
+    /* Process and return the data. */
+    dataset = collection_to_mvm_objects(tc, tc->instance->heap_snapshots);
     destroy_heap_snapshot_collection(tc);
     return dataset;
 }
